#include "feedback.h"
#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gdk/gdkx.h>
#include <gdk/gdkprivate.h>
#include <stdio.h>
#include <string.h>
#include <orb/orbit.h>
#include <gnome.h>
#include <libgnorba/gnorba.h>
#include <liboaf/liboaf.h>
#include <sys/types.h>
#include <unistd.h>
#include <X11/Xmu/WinUtil.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <dirent.h>

Feedback feedback_client;
static gboolean set_launch_env(char *);

int main (int argc, char *argv[])
{
	CORBA_Environment ev;
	CORBA_ORB orb;
	gchar **argv_copy, *progname, *prog, desktop_file[BUFSIZ], *tmp=NULL, *dirname, launch_id[BUFSIZ+1], *ior, buffer[BUFSIZ+1], *install_dir=NULL;
	gint pid, i;
	gboolean complyFlag=FALSE;
	static const guint prop_buffer_length = 1024 * 1024;
	Atom xa_win_prop=0, actual_type_return=0; 
	gint actual_format_return=0;
	Window *prop_return=NULL, root_window=0;
	unsigned long nitems_return=0, bytes_after_return=0;
	Display *dpy;
	Feedback_WindowList mapped_windows;
	DIR *dp1, *dp2;
	struct dirent *d1, *d2;
	FILE *ifp;


	if (argc != 2){
		fprintf(stderr, "Wrong no. of arguments\n");
		gtk_exit(1);
	}
	argv_copy=g_new(char *, argc+1);
	for(i=0; i<argc; i++) {
		argv_copy[i]=argv[i];
	}
        
	/* get name of app. to be started */
	prog=argv[1];
	if((tmp=strrchr(prog, '/'))!=NULL){
		progname=tmp++;
	}else
		progname=prog;
	sprintf(desktop_file, "%s.desktop", progname);

	CORBA_exception_init(&ev);
	/*gnome_init_with_popt_table("feedback-client", "0.0", argc,
					argv, oaf_popt_options, 0, NULL);

	oaf_init(argc, argv);*/

	orb = gnome_CORBA_init("Feedback", "v 1.0", &argc, argv, 0, &ev);

	
	 ifp = fopen("feedback.ior","r");
        if( ifp == NULL ) {
			fprintf(stderr, "No feedback.ior file!");
				gtk_exit(-1);
        }

	fgets(buffer,BUFSIZ,ifp);
        ior = g_strdup(buffer);

	fclose(ifp);


       /*feedback_client=oaf_activate_from_id("OAFIID:Feedback:afd9e479-24b6-4c24-8945-168da0cd01b0", 0, NULL, &ev);*/

	feedback_client=CORBA_ORB_string_to_object(orb, ior, &ev);

	if (!feedback_client) {
		fprintf(stderr, "Error: can't bind\n");
		gtk_exit(1);
	}

	dpy=(Display *)GDK_DISPLAY();
	root_window=GDK_WINDOW_XWINDOW(gdk_window_lookup(GDK_ROOT_WINDOW()));

        /* set DESKTOP_LAUNCH_ID env. variable */
	if (set_launch_env(launch_id) == FALSE){
		fprintf(stderr, "Can't set environmental variable\n");
		gtk_exit(-1);
	}

	/* hack to see if app to be started is gnome, ie checks for .desktop file.
	   If it is a gnome app, then the compliance flag (complyFlag) is set to TRUE*/ 
	if(getenv("INSTALL_DIR"))
		install_dir=g_strconcat(getenv("INSTALL_DIR"),"/share/gnome/apps/", NULL);
        else
		install_dir=g_strdup("/usr/share/gnome/apps/");
		
	if ((dp1 = opendir(install_dir)) == NULL) {
		fprintf(stderr, "Can't open desktop directories\n");
		complyFlag=FALSE;
	}else {
		while ((d1=readdir(dp1)) != NULL){
			if(d1->d_name[0] == '.')
				continue;
			dirname=g_malloc(strlen(install_dir) + strlen(d1->d_name) +3);
			sprintf(dirname, "%s/%s", install_dir, d1->d_name);
			if((dp2 = opendir(dirname)) != NULL) {	
				while ((d2=readdir(dp2)) != NULL){
					if(d2->d_name[0] == '.')
						continue;
					if(strcmp(d2->d_name, desktop_file)== 0){
						complyFlag=TRUE;
					}
				}	
			        closedir(dp2);
				g_free(dirname);
			}
			if(strcmp(d1->d_name, desktop_file)==0){
				complyFlag=TRUE;
			}
		}	
		closedir(dp1);
	}

	/* Get the current value of the root window property, _WIN_CLIENT_LIST.
	   Extract the list of windows, assign these to an array which will be 
 	   passed to the remote method monitorWindows() */
	xa_win_prop=XInternAtom(dpy, "_WIN_CLIENT_LIST", FALSE);
  	XGetWindowProperty(dpy,
			root_window,
			xa_win_prop,
			0L,
              		prop_buffer_length,
			False, XA_CARDINAL,
			&actual_type_return,
			&actual_format_return,
			&nitems_return,
			&bytes_after_return,
			(unsigned char **)(&prop_return));

	mapped_windows._length=nitems_return;
	mapped_windows._buffer=(CORBA_long *)prop_return;
	for (i=1; i < nitems_return; i++){
		mapped_windows._buffer[i]=(CORBA_long )prop_return[i];

	}
        /* call the remote method to start the visual feedback */
	Feedback_startFeedback(feedback_client, launch_id, progname, &ev);
	if (ev._major != CORBA_NO_EXCEPTION) {
			fprintf(stderr, "exception from startFeedback\n");
			gtk_exit(-1);
        }

	/* call the remote method to monitor the _WIN_CLIENT_LIST property
	   on the root window.  This is a oneway remote method*/
	Feedback_monitorWindows(feedback_client, launch_id, progname, &mapped_windows, complyFlag, &ev);
	/* fork and exec the application */
	 switch (pid = fork())
        {
			case -1:
				fprintf (stderr, "Error forking\n");
				gtk_exit(1);
			case 0:
				
				execvp (prog, argv_copy+1);
				fprintf(stderr, "Error executing program\n");
	 			Feedback_stopFeedback(feedback_client, launch_id, progname, &ev);
				_exit(-1);
	}

	g_free(ior);
	g_free(install_dir);
	CORBA_Object_release(feedback_client, &ev);
	CORBA_Object_release((CORBA_Object)orb, &ev);
	return(0); 
}	

static gboolean set_launch_env(char *launch_id)
{
	static gchar launch_env[BUFSIZ+1];
	gchar *machine_name=NULL, *display_name=NULL, hostname[BUFSIZ+1];
	if(gethostname(hostname, BUFSIZ)==0)
		machine_name=g_strdup(hostname);
	display_name=getenv("DISPLAY");
	if (machine_name!=NULL && display_name!=NULL){
		snprintf(launch_id, BUFSIZ-20, "%s %s %d", machine_name, display_name, (int)getpid());
		snprintf(launch_env, BUFSIZ, "DESKTOP_LAUNCH_ID=%s", launch_id); 	
		putenv(launch_env);
		return TRUE;
	}else
		return FALSE;
}
	

	
	
