#include <gtk/gtk.h>
#include <gdk/gdk.h>
#include <gdk/gdkx.h>
#include <gdk/gdkprivate.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <string.h>
#include "feedback.h"
#include "hourglass.xpm"

/*** App-specific servant structures ***/
typedef struct {
POA_Feedback servant;
PortableServer_POA poa;




} impl_POA_Feedback;
/* struct for holding values necessary for root_event_monitor() */
typedef struct {
gchar *launch_info;
gchar *prog;
GSList *window_list;
gboolean comply;
CORBA_Environment *ev;
impl_POA_Feedback *servant;
}monitorInfo;

static monitorInfo info;

/*** Implementation stub prototypes ***/
static void impl_Feedback__destroy(impl_POA_Feedback *servant,
CORBA_Environment *ev);

static void
impl_Feedback_startFeedback(impl_POA_Feedback *servant,
CORBA_char * launchInfo,
CORBA_char * progName,
CORBA_Environment *ev);

static void
impl_Feedback_monitorWindows(impl_POA_Feedback *servant,
CORBA_char * launchInfo,
CORBA_char * progName,
Feedback_WindowList* list,
CORBA_boolean complyFlag,
CORBA_Environment *ev);

static void
impl_Feedback_stopFeedback(impl_POA_Feedback *servant,
CORBA_char * launchInfo,
CORBA_char * progName,
CORBA_Environment *ev);
static void
create_invisible(gchar *title);
static void
set_icon(GdkWindow *window);

static GdkFilterReturn	rootwin_event_monitor	  (GdkXEvent		*gdk_xevent,
                                                   GdkEvent		*event,
                                                   gpointer             data);


/*** epv structures ***/
static PortableServer_ServantBase__epv impl_Feedback_base_epv = {
NULL, /* _private data */
NULL, /* finalize routine */
NULL, /* default_POA routine */
};
static POA_Feedback__epv impl_Feedback_epv = {
NULL, /* _private */

(gpointer)&impl_Feedback_startFeedback,

(gpointer)&impl_Feedback_monitorWindows,

(gpointer)&impl_Feedback_stopFeedback,

};

/*** vepv structures ***/
static POA_Feedback__vepv impl_Feedback_vepv = {
&impl_Feedback_base_epv,
&impl_Feedback_epv,
};

/*** Stub implementations ***/
static Feedback impl_Feedback__create(PortableServer_POA poa, CORBA_Environment *ev)
{
Feedback retval;
impl_POA_Feedback *newservant;
PortableServer_ObjectId *objid;

newservant = g_new0(impl_POA_Feedback, 1);
newservant->servant.vepv = &impl_Feedback_vepv;
newservant->poa = poa;
POA_Feedback__init((PortableServer_Servant)newservant, ev);
objid = PortableServer_POA_activate_object(poa, newservant, ev);
CORBA_free(objid);
retval = PortableServer_POA_servant_to_reference(poa, newservant, ev);

return retval;
}

static void
impl_Feedback__destroy(impl_POA_Feedback *servant, CORBA_Environment *ev)
{
PortableServer_ObjectId *objid;

objid = PortableServer_POA_servant_to_id(servant->poa, servant, ev);
PortableServer_POA_deactivate_object(servant->poa, objid, ev);
CORBA_free(objid);

POA_Feedback__fini((PortableServer_Servant)servant, ev);
g_free(servant);
}


static void
impl_Feedback_startFeedback(impl_POA_Feedback *servant,
CORBA_char * launchInfo,
CORBA_char * progName,
CORBA_Environment *ev)
{
	/* the create_invisible function is hacked from the 
	   Xalf code.  The plan is to create a plug-in interface to
	   provide for different types of feedback */
	create_invisible((gchar *)progName);
}


static void
impl_Feedback_monitorWindows(impl_POA_Feedback *servant,
CORBA_char * launchInfo,
CORBA_char * progName,
Feedback_WindowList* list,
CORBA_boolean complyFlag,
CORBA_Environment *ev)
{
	GdkWindow *root_win;
	XWindowAttributes root_attribs;
	gint i, win_num;
	root_win=gdk_window_lookup(GDK_ROOT_WINDOW());
	XGetWindowAttributes(GDK_DISPLAY(),
				GDK_ROOT_WINDOW(),
				&root_attribs);
	/* monitor for change property on root window */
	XSelectInput(GDK_DISPLAY(),
			 GDK_ROOT_WINDOW(),
                         root_attribs.your_event_mask | PropertyChangeMask);
	/* assign required values to global monitorInfo struct variable */
	info.comply=(gboolean)complyFlag;
	info.ev=ev;	
	info.servant=servant;
	info.launch_info=g_strdup(launchInfo);
	info.prog=g_strdup(progName);
	for(i=0; i<list->_length; i++) {
		win_num=(gint)list->_buffer[i];
		info.window_list=g_slist_prepend(info.window_list, GINT_TO_POINTER(win_num)); 
	}
	gdk_window_add_filter (root_win,
                               rootwin_event_monitor,
				NULL);
}


static void
impl_Feedback_stopFeedback(impl_POA_Feedback *servant,
CORBA_char * launchInfo,
CORBA_char * progName,
CORBA_Environment *ev)
{
	gtk_exit(0);
}


void
set_icon (GdkWindow *window)
{
    static GdkPixmap *w_minipixmap = NULL;
    static GdkBitmap *w_minimask = NULL;
    GdkAtom icon_atom;
    glong data[2];

    w_minipixmap =
	gdk_pixmap_create_from_xpm_d (window,
                                      &w_minimask,
                                      NULL,
                                      hourglass_xpm);

    data[0] = ((GdkPixmapPrivate *)w_minipixmap)->xwindow;
    data[1] = ((GdkPixmapPrivate *)w_minimask)->xwindow;

    icon_atom = gdk_atom_intern ("KWM_WIN_ICON", FALSE);
    gdk_property_change (window, icon_atom, icon_atom,
			 32, GDK_PROP_MODE_REPLACE,
			 (guchar *)data, 2);
}

static void create_invisible(gchar *title)
{
    GtkWidget *window;
    gchar *tasktitle;

    tasktitle = g_strconcat ("(", title, ")", NULL);
    /* Create the indicator: An invisible window, to be shown in a pager */
    window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title (GTK_WINDOW(window), tasktitle);
    gtk_window_set_policy (GTK_WINDOW(window), FALSE, FALSE, TRUE);
    gtk_window_set_wmclass (GTK_WINDOW(window), "feedback", "feedback");
    gtk_widget_realize (window);
    gdk_window_set_decorations (window->window, 0);
    /* Set a hourglass icon for the indicator via KWM_WIN_ICON */
    set_icon (window->window);

    /* Show window */
    gtk_widget_show (window);
}

static GdkFilterReturn
rootwin_event_monitor (GdkXEvent *gdk_xevent,
                    GdkEvent  *event,
                    gpointer   data)
{
   static const guint prop_buffer_length = 1024 * 1024;
   Atom xa_win_prop=0, win_actual_type_return=0, launch_actual_type_return=0, class_actual_type_return=0, xa_launch_prop=0, xa_class_prop=0;
   gint i, win_num, win_actual_format_return=0, launch_actual_format_return=0, class_actual_format_return=0;
   Window *win_prop_return, root_window;
   Display *dpy;
   gchar *launch_prop_return=NULL, *class_prop_return=NULL;
   unsigned long win_nitems_return=0, launch_nitems_return=0, class_nitems_return=0, win_bytes_after_return=0, launch_bytes_after_return=0, class_bytes_after_return=0;
   XEvent *xevent = gdk_xevent;

   dpy=(Display *)GDK_DISPLAY();
   root_window=GDK_WINDOW_XWINDOW(gdk_window_lookup(GDK_ROOT_WINDOW()));
   xa_win_prop=XInternAtom(dpy, "_WIN_CLIENT_LIST", FALSE);
   xa_launch_prop=XInternAtom(dpy, "_NET_WM_LAUNCH_ID", FALSE);

   switch (xevent->type)
   {
      case PropertyNotify:
         if(xevent->xproperty.atom == xa_win_prop){
	    /* find the current value of changed _WIN_CLIENT_LIST property */
     	    XGetWindowProperty(dpy,
			root_window,
			xa_win_prop,
			0L,
			prop_buffer_length,
			False, XA_CARDINAL,
			&win_actual_type_return,
			&win_actual_format_return,
			&win_nitems_return,
			&win_bytes_after_return,
			(unsigned char **)(&win_prop_return));

            for(i=(win_nitems_return-1); i>=0; i--){
	        /* find the new window, ie not in list of windows included in
		  parameter to monitorWindows */
	       win_num=(gint)win_prop_return[i];
	       if(!g_slist_find(info.window_list, GINT_TO_POINTER(win_num))){
                /*if the application sets _NET_WM_LAUNCH_PID */
	          XGetWindowProperty(dpy,
                                     win_prop_return[i],
                                     xa_launch_prop,
                                     0L,
                                     prop_buffer_length,
                                     False, XA_STRING,
                                     &launch_actual_type_return,
                                     &launch_actual_format_return,
                                     &launch_nitems_return,
                                     &launch_bytes_after_return,
                                    (unsigned char **)(&launch_prop_return));
		  /* if _NET_WM_LAUNCH_ID property is set on the new window, see
		     if the property value is correct */
		  if(launch_prop_return){
 		     if(strcmp((char *)info.launch_info, launch_prop_return)==0) {
                        impl_Feedback_stopFeedback(info.servant, info.launch_info, info.prog, info.ev);
                     }
                  }else if (info.comply==FALSE){  /*this code is only entered if not sure
						   whether or not new app supports
						   _NET_WM_LAUNCH_PID */
                    	xa_class_prop=XInternAtom(dpy, "WM_CLASS", FALSE);
                        XGetWindowProperty(dpy,
                                       	win_prop_return[i],
                                      	xa_class_prop,
                                       	0L,
                                      	prop_buffer_length,
                                      	False, XA_STRING,
                                      	&class_actual_type_return,
                                      	&class_actual_format_return,
                                      	&class_nitems_return,
                                      	&class_bytes_after_return,
                                      	(unsigned char **)(&class_prop_return));
                         if(class_prop_return){
                            if(strstr(class_prop_return, "feedback")==NULL){
                               impl_Feedback_stopFeedback(info.servant, info.launch_info, info.prog, info.ev);
                            }
                         }else{
                            impl_Feedback_stopFeedback(info.servant, info.launch_info, info.prog, info.ev);
			}
                   }
               }
            }  
         }
         break;
     default:
	 break;
   }
  return GDK_FILTER_CONTINUE;
}

